#include "h/types.h"
#include <stdio.h>
#include "h/emuGeneral.h"
#include "../nes/h/nesGeneral.h"
#include "../nes/h/cpu.h"
#include "../nes/h/ppu.h"
#include "../nes/blargg/Nes_Apu.h"
#include "../nes/blargg/Nes_Apu_Reflector.h"
#include "../nes/h/apu.h"

#include "h/interface.h"
#include "h/mapper.h"

namespace Reminesce
{

u8 StateNum;

/*
	10 ram
	11 cpu registers
	
	20 ppu registers
	21 sprram
	22 palette

	40 input regs
	50 wram size
	51 wram
	52 nametable flags
	53 nametable 0
	54 nametable 1
	55 chrram

	60 APU error
	70 Mapper error
*/

int emuSaveState(FILE *f)
{
	u8 *arr; unsigned int ret, i;
	/*						CPU							*/
	// $800
	if (fwrite(ram, 1, 0x800, f) != 0x800)						//Ram
		return 10;
	
	// $7
	arr = new u8 [7];											//CPU Registers
	arr[0] = (cpu.N & cN) | (cpu.V ? cV : 0) | 0 | 0 | cpu.D | cpu.I | (cpu.Z ? 0: cZ) | cpu.C;
	arr[1] = cpu.a;
	arr[2] = cpu.x;
	arr[3] = cpu.y;
	arr[4] = cpu.s;
	arr[5] = (u8)cpu.pc;
	arr[6] = cpu.pc>>8;

	ret = fwrite(arr, 1, 7, f);
	delete [] arr; arr = 0;
	if (ret != 0x7)
		return 11;
	
	/*						PPU							*/
	// 10
	arr = new u8 [10];											//PPU & its regs
	arr[0] = ppu.r2000;
	arr[1] = ppu.r2001;
	arr[2] = ppu.sprLoc;
	arr[3] = (u8)ppu.temp;
	arr[4] = ppu.temp>>8;
	arr[5] = (u8)ppu.address;
	arr[6] = ppu.address>>8;
	arr[7] = ppu.pipe;
	arr[8] = ppu.bus;
	arr[9] = ((ppu.secondWrite)	?0x8 :0) | (ppu.xLatch & 7);

	ret = fwrite(arr, 1, 10, f);								//write ppu registers
	delete [] arr; arr = 0;
	if (ret != 10)
		return 20;

	ret = fwrite(sprRam, 1, 256, f);							//spr ram
	if (ret != 256)
		return 21;

	//Palette
	ret = fwrite(pal, 1, 0x20, f);								//palette
	if (ret != 0x20)
		return 22;

	/*						Input						*/
	arr = new u8 [4];
	arr[0] = (ButtonShift1 & 0x7F) | (Strobing ? 0x80 : 0);
	arr[1] = (ButtonShift2 & 0x7F);
	arr[2] = 0;		//Will hold input device input1
	arr[3] = 0;		//Will hold input device input2
	
	ret = fwrite(arr, 1, 4, f);									//write ppu registers
	delete [] arr; arr = 0;
	if (ret != 4)
		return 40;

	/*						misc						*/
	
	//wram
	if (sramSize == 0)
	{
		const u16 wramDefaultSize = 0x2000;
		for (i=0; i<wramDefaultSize; i++)
		{
			if (wram[i])
				break;
		}
		if (i == wramDefaultSize)
			i = 0;
		else
			i = wramDefaultSize;
	}
	else
	{
		for (i=0; i<sramSize; i++)
		{
			if (wram[i])
				break;
		}
		if (i == sramSize)
			i = 0;
		else
			i = sramSize;
	}

	if (fwrite(&i, 1, 2, f) != 2)
		return 50;

	if (i)
	{
		if (fwrite(wram, 1, i, f) != i)
			return 51;
	}

	//nametable 0
	for (i=0; i< 0x400; i++)
	{
		if (nameTable[i])
			break;
	}
	ret = (i != 0x400);
	
	//nametable 1
	for (i=0x400; i< 0x800; i++)
	{
		if (nameTable[i])
			break;
	}
	ret |= (i != 0x800)<<1;

	//chrram
	for (i=0; i < 0x2000; i++)
	{
		if (chrRam[i])
			break;
	}
	ret |= (i != 0x2000)<<2;

	i = cpuCycles<<4 | ret;			//should only be 3 bits since longest opcode can use
									//7 cycles, 4 bits is availible but 3 is enough
									//it should never be 7 or more cycles anyways.
	//nametable write flags
	if (fwrite(&i, 1, 1, f) != 1)
		return 52;
	
	//nametable 0 write
	if (i & 1)
	{
		if (fwrite(&nameTable[0x000], 1, 0x400, f) != 0x400)
			return 53;
	}
	
	//nametable 1 write
	if (i & 2)
	{
		if (fwrite(&nameTable[0x400], 1, 0x400, f) != 0x400)
			return 54;
	}

	//chrram
	if (i & 4)
	{
		if (fwrite(chrRam, 1, chrramSize, f) != chrramSize)
			return 55;
	}

	/*						APU							*/
	Nes_Apu_Reflector::save(apu, f);

	if (mapInfo->SaveMI)
	{
		arr = new u8 [128];
		mapInfo->SaveMI(arr);
		ret = fwrite(arr, 1, 128, f);
		delete [] arr;

		if (ret != 128)
			return 70;
	}

	return 0;
}

int emuSizeState(u32 &size)
{
	unsigned int ret, i;

	size = 0x800;
	size += 0x7;
	size += 0x10;
	size += 256;

	size += 0x20;
	size += 0x4;

	//wram
	if (sramSize == 0)
	{
		const u16 wramDefaultSize = 0x2000;
		for (i=0; i<wramDefaultSize; i++)
		{
			if (wram[i])
				break;
		}
		if (i == wramDefaultSize)
			i = 0;
		else
			i = wramDefaultSize;
	}
	else
	{
		for (i=0; i<sramSize; i++)
		{
			if (wram[i])
				break;
		}
		if (i == sramSize)
			i = 0;
		else
			i = sramSize;
	}

	size += 0x2;

	if (i)
	{
		size += i;
	}

	//nametable 0
	for (i=0; i< 0x400; i++)
	{
		if (nameTable[i])
			break;
	}
	ret = (i != 0x400);
	
	//nametable 1
	for (i=0x400; i< 0x800; i++)
	{
		if (nameTable[i])
			break;
	}
	ret |= (i != 0x800)<<1;

	//chrram
	for (i=0; i < 0x2000; i++)
	{
		if (chrRam[i])
			break;
	}
	ret |= (i != 0x2000)<<2;

	i = cpuCycles<<4 | ret;			//should only be 3 bits since longest opcode can use
									//7 cycles, 4 bits is availible but 3 is enough
									//it should never be 7 or more cycles anyways.

	size += 1;

	//nametable 0 write
	if (i & 1)
	{
		size += 0x400;
	}
	
	//nametable 1 write
	if (i & 2)
	{
		size += 0x400;
	}

	//chrram
	if (i & 4)
	{
		size += chrramSize;
	}

	/*						APU							*/
//	Nes_Apu_Reflector::size(apu, size);

	if (mapInfo->SaveMI)
	{
		size += 128;
	}

	return 0;
}



int emuLoadState(FILE *f)
{
	u8 *arr; int ret, i;
	/*						CPU							*/
	// $800
	if ((i = fread(ram, 1, 0x800, f)) != 0x800)						//Ram
		return 10;
	
	// $7
	arr = new u8 [7];											//CPU Registers
	if (fread(arr, 1, 7, f) != 7)
		{ ret = 11; goto errHdl;}
	
	cpu.N = arr[0] & cN;
	cpu.V = arr[0] & cV;
	// 0
	// 0
	cpu.D = arr[0] & cD;
	cpu.I = arr[0] & cI;
	cpu.Z = !(arr[0] & cZ);
	cpu.C = arr[0] & cC;

	cpu.a = arr[1];
	cpu.x = arr[2];
	cpu.y = arr[3];
	cpu.s = arr[4];
	cpu.pc= arr[5] | (arr[6]<<8);

	delete [] arr; arr = 0;

	
	/*						PPU							*/
	// 10
	arr = new u8 [10];											//PPU & its regs
	if (fread(arr, 1, 10, f) != 10)								//read ppu registers
		{ ret = 20; goto errHdl;}

	ppu.r2000		= arr[0];
	ppu.r2001		= arr[1];
	ppu.sprLoc		= arr[2];
	ppu.temp		= arr[3] | arr[4]<<8;
	ppu.address		= arr[5] | arr[6]<<8;
	ppu.pipe		= arr[7];
	ppu.bus			= arr[8];

	ppu.secondWrite = (arr[9] & 8) != 0;
	ppu.xLatch		= arr[9] & 7;
	
	delete [] arr; arr = 0;

	if (fread(sprRam, 1, 256, f) != 256)						//spr ram
		return 21;

	//Palette
	if (fread(pal, 1, 0x20, f) != 0x20)							//palette
		return 22;

	/*						Input						*/
	arr = new u8 [4];
	if (fread(arr, 1, 4, f) != 4)								//write ppu registers
		{ ret = 40; goto errHdl;}
	ButtonShift1 = arr[0] & 0x7F;
	Strobing = (arr[0] & 0x80) ? 1 : 0;
	ButtonShift2 = arr[1] & 0x7F;
	arr[2] = 0;		//Will hold input device input1
	arr[3] = 0;		//Will hold input device input2
	
	delete [] arr; arr = 0;

	/*						misc						*/
	i=0;
	if (fread(&i, 1, 2, f) != 2)
		return 50;

	if (i)
	{
		if (fread(wram, 1, i, f) != (unsigned)i)
			return 51;
	}

	//nametable write flags
	if (fread(&i, 1, 1, f) != 1)
		return 52;

	cpuCycles = (i>>4) & 7;

	//nametable write 1
	if (i & 1)
	{
		if (fread(&nameTable[0x000], 1, 0x400, f) != 0x400)
			return 53;
	}
	else
		memset(&nameTable[0x000], 0, 0x400);

	//nametable write 2
	if (i & 2)
	{
		if (fread(&nameTable[0x400], 1, 0x400, f) != 0x400)
			return 54;
	}
	else
		memset(&nameTable[0x400], 0, 0x400);

	//chrram
	if (i & 4)
	{
		if (fread(chrRam, 1, 0x2000, f) != 0x2000)
			return 55;
	}

	/*						APU							*/
	resetTheApu();
	Nes_Apu_Reflector::load(f, apu);

	if (mapInfo->LoadMI)
	{
		arr = new u8 [128];
		ret = fread(arr, 1, 128, f);
		if (ret != 128)
			{ ret = 70; goto errHdl;}

		mapInfo->LoadMI(arr);
	}
	ret = 0;
errHdl:
	delete [] arr;
	return ret;
}

}
