#include "h/types.h"
#include <string.h>
#include "h/interface.h"
#include "h/mapper.h"
#include "../nes/h/cpu.h"
#include "../nes/h/nes.h"
#include "../nes/h/ppu.h"
#include "../nes/h/nesGeneral.h"
#include "h/emuFileHandling.h"
#include "../nes/h/registers.h"
#include "../nes/h/interruptions.h"

namespace Reminesce
{

MapperParam mp;
PDLLInfo nestenDll;
CPMapperInfo mapInfo = 0;

typedef PDLLInfo	(*nestenLoad_DLL)	(int VersionRequired);
typedef void		(*nestenUnload_DLL)	();

nestenLoad_DLL		nestenMapLoad;
nestenUnload_DLL	nestenMapUnload;

unsigned int sramSize;

typedef	int		(__cdecl *PReadFunc)	(int bank, int where);
typedef	void	(__cdecl *PWriteFunc)	(int bank, int where, int what);

/*				read write func					*/
PWriteFunc pWriteFunc[0x10];
PReadFunc  pReadFunc[0x10];

void setbankWritePtr(int bank, PWriteFunc pFunc)
{
	pWriteFunc[bank] = pFunc;
}

PWriteFunc getbankWritePtr(int bank)
{
	return pWriteFunc[bank];
}

void setbankReadPtr(int bank, PReadFunc pFunc)
{
	pReadFunc[bank] = pFunc;
}

PReadFunc getbankReadPtr(int bank)
{
	return pReadFunc[bank];
}

/*				Prg switching				*/
void  mySetPRG_ROM4  (int bank, int where) { setPrgBanks(bank<<12,  where*4,  4);  }
void  mySetPRG_ROM8  (int bank, int where) { setPrgBanks(bank<<12,  where*8,  8);  }
void  mySetPRG_ROM16 (int bank, int where) { setPrgBanks(bank<<12,  where*16, 16); }
void  mySetPRG_ROM32 (int bank, int where) { setPrgBanks(bank<<12,  where*32, 32); }

void  mySetPRG_RAM4  (int bank, int where) { setPrgBanks2(bank<<12, where*4,  4);  }
void  mySetPRG_RAM8  (int bank, int where) { setPrgBanks2(bank<<12, where*8,  8);  }
void  mySetPRG_RAM16 (int bank, int where) { setPrgBanks2(bank<<12, where*16, 16); }
void  mySetPRG_RAM32 (int bank, int where) { setPrgBanks2(bank<<12, where*32, 32); }

/*				Chr switching				*/
void  mySetCHR_ROM1  (int bank, int where) { setChrBanks(bank<<10,  where*1, 1); }
void  mySetCHR_ROM2  (int bank, int where) { setChrBanks(bank<<10,  where*2, 2); }
void  mySetCHR_ROM4  (int bank, int where) { setChrBanks(bank<<10,  where*4, 4); }
void  mySetCHR_ROM8  (int bank, int where) { setChrBanks(bank<<10,  where*8, 8); }

void  mySetCHR_RAM1  (int bank, int where) { setChrBanks2(bank<<10, where*1, 1); }
void  mySetCHR_RAM2  (int bank, int where) { setChrBanks2(bank<<10, where*2, 2); }
void  mySetCHR_RAM4  (int bank, int where) { setChrBanks2(bank<<10, where*4, 4); }
void  mySetCHR_RAM8  (int bank, int where) { setChrBanks2(bank<<10, where*8, 8); }

/*				Mirroring					*/
void setMirror_H()  { setMirror(horz); }
void setMirror_V()  { setMirror(vert); }
void setMirror_4()  { setMirror(fScreen); }
void setMirror_S0() { setMirror(single0); }
void setMirror_S1() { setMirror(single1); }

void setMirror_Custom(int a, int b, int c, int d)
{
	pntBank[0] = &nameTable[0x400 * a];
	pntBank[1] = &nameTable[0x400 * b];
	pntBank[2] = &nameTable[0x400 * c];
	pntBank[3] = &nameTable[0x400 * d];
}

void  SetPRG_OB4 (int bank)
{
	pbankRead[ bank<<1    ] = &openBanks[  (bank-4)<<12];
	pbankRead[(bank<<1) +1] = &openBanks[(((bank-4)<<1) + 1)<<11];

	pbankWrite[bank<<1] = pbankWrite[bank<<1 + 1] = deadWrite;
}

char *GetPRG_Ptr4(int bank)
{
	//! Not sure if this is correct
	return (char*)pbankRead[bank<<1];
}

char *GetCHR_Ptr1(int bank) { 	return (char*)pChrRead[bank]; }

int GetPRG_ROM4(int bank)
{
	int ptr = pbankRead[bank<<1] - prgromPtr;
	if ((ptr < 0) || (ptr > header.prg * 16 * 1024))
		return -1;
	else
		return ptr / (1024 * 4);
}

int GetPRG_RAM4(int bank)
{
	int ptr = pbankWrite[bank<<1] - wram;
	if ((ptr < 0) || (ptr > header.prg * 16 * 1024))
		return -1;
	else
		return ptr / (1024 * 4);
}

int GetCHR_ROM1(int bank)	{ return (pChrWrite[bank] - chrromPtr)/1024;}
int GetCHR_RAM1(int bank)	{ return (pChrRead [bank] - chrRam)/1024; }

void bankWrite(int val, int Addr)
{
	pWriteFunc[Addr>>12](Addr>>12, Addr & 0x0FFF, val);
}

int bankRead(int Addr)
{
	return pReadFunc[Addr>>12](Addr>>12, Addr & 0x0FFF);
}


//Sram
void Set_SramSize(int size)
{
	if (size > wramSize)
	{
		emuMsg("sram size to big");
	}
	sramSize = size;
}

void Save_SRAM()
{
	SaveWram();
}

void Load_SRAM()
{
	LoadWram();
}

void DbgOut(char *p)
{
}
void StatusOut(char *p)
{
}

//not done
int GetMenuRoot()
{
	return 0;
}

int AddMenuItem(int,char *,int,int,int,int,int)
{
	return 0;
}

void MMC5_UpdateAttributeCache()
{
}


void initMp()
{
	memset(&mp, 0, sizeof(mp));

	mp.SetPRG_ROM4		= mySetPRG_ROM4;
	mp.SetPRG_ROM8		= mySetPRG_ROM8;
	mp.SetPRG_ROM16		= mySetPRG_ROM16;
	mp.SetPRG_ROM32		= mySetPRG_ROM32;

	mp.SetPRG_RAM4		= mySetPRG_RAM4;
	mp.SetPRG_RAM8		= mySetPRG_RAM8;
	mp.SetPRG_RAM16		= mySetPRG_RAM16;
	mp.SetPRG_RAM32		= mySetPRG_RAM32;

	mp.SetCHR_ROM1		= mySetCHR_ROM1;
	mp.SetCHR_ROM2		= mySetCHR_ROM2;
	mp.SetCHR_ROM4		= mySetCHR_ROM4;
	mp.SetCHR_ROM8		= mySetCHR_ROM8;

	mp.SetCHR_RAM1		= mySetCHR_RAM1;
	mp.SetCHR_RAM2		= mySetCHR_RAM2;
	mp.SetCHR_RAM4		= mySetCHR_RAM4;
	mp.SetCHR_RAM8		= mySetCHR_RAM8;

	mp.SetWriteHandler	= setbankWritePtr;
	mp.GetWriteHandler	= getbankWritePtr;
	mp.SetReadHandler	= setbankReadPtr;
	mp.GetReadHandler	= getbankReadPtr;
	
	mp.Mirror_H			= setMirror_H;
	mp.Mirror_V			= setMirror_V;
	mp.Mirror_4			= setMirror_4;
	mp.Mirror_S0		= setMirror_S0;
	mp.Mirror_S1		= setMirror_S1;
	mp.Mirror_Custom	= setMirror_Custom;

	mp.SetPRG_OB4		= SetPRG_OB4;
	mp.GetPRG_Ptr4 		= GetPRG_Ptr4;
	mp.IRQ				= IRQ;

	//not done
	mp.Set_SRAMSize		= Set_SramSize;
	mp.Save_SRAM		= Save_SRAM;
	mp.Load_SRAM		= Load_SRAM;

	mp.DbgOut			= DbgOut;
	mp.StatusOut		= StatusOut;

	mp.GetPRG_ROM4		= GetPRG_ROM4;
	mp.GetPRG_RAM4		= GetPRG_RAM4;
	mp.GetCHR_ROM1		= GetCHR_ROM1;
	mp.GetCHR_RAM1		= GetCHR_RAM1;
	mp.GetCHR_Ptr1		= GetCHR_Ptr1;
	mp.GetMenuRoot		= GetMenuRoot;
	mp.AddMenuItem		= AddMenuItem;

	mp.MMC5_UpdateAttributeCache	= MMC5_UpdateAttributeCache;
}

char loadMapperNumber(int mapNum)
{
	mapInfo = nestenDll->LoadMapper(mapNum);
	if (!mapInfo)
		return 1;

	mp.PRG_ROM_Size = header.prg * 0x400 * 16;
	mp.CHR_ROM_Size = header.chr * 0x400 * 8;
	mp.Flags = header.fourScreenMirroring | header.vmirror | header.sram;

	return 0;
}

}
