namespace Reminesce
{

void inline BRANCH(u8 b);
//if true is not 1, C will not work correctly

inline u16 getMode(u8 val, u8 write)
{
	switch (val)
	{
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
			return 0;
		case 5:
			return write ? logged_Absolute1w : logged_Absolute1r;
		case 6:
			return write ? logged_AbsoluteXw : logged_AbsoluteXr;
		case 7:
			return write ? logged_AbsoluteYw : logged_AbsoluteYr;
		case 8:
			return write ? logged_AbsoluteXw : logged_AbsoluteXr;
		case 9:
			return write ? logged_AbsoluteYw : logged_AbsoluteYr;
		case 10:
			return write ? logged_indirectXw : logged_indirectXr;
		case 12:
			return write ? logged_indirectYw : logged_indirectYr;
			break;
	}
	return 0;
}

#undef loggedOffset
#define loggedOffset(loc) (pbankRead[(loc)>>11] - prgromPtr + ((loc) & 0x7FF))

void inline ADC()
{
	tempVar temp1, temp2;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	temp1.si_		= Read6502();
	temp2.si_		= (cpu.a + temp1.si_ + cpu.C);
	cpu.V			= ((!((cpu.a ^ temp1.si_) & 0x80)) && ((cpu.a ^ temp2.si_) & 0x80));
	cpu.C			= temp2.si_ > 0xff;
	cpu.N			= cpu.Z = cpu.a = temp2.t16_.u.l;
}

void inline AND()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	cpu.N = cpu.Z = cpu.a &= Read6502();
}

void inline ASL()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
		{
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
		}
	#endif
	temp1.t16_.u.l = Read6502();
	Write6502(temp1.t16_.u.l);
	cpu.C  = (temp1.t16_.u.l & 0x80) > 0;
	cpu.N  = cpu.Z = (temp1.t16_.u.l << 1);
	Write6502(temp1.t16_.u.l << 1);
}

void inline ASLA()
{
	cpu.C  = (cpu.a & 0x80) > 0;
	cpu.N = cpu.Z = cpu.a = (u8)(cpu.a<<1);
}

void inline BCC()
{
	BRANCH(!cpu.C);
}

void inline BCS()
{
	BRANCH(cpu.C);
}

void inline BEQ()
{
	BRANCH(!cpu.Z);
}

void inline BIT()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	temp1.si_   = Read6502();
	cpu.N  = temp1.si_  & cN;
	cpu.V  = temp1.si_  & cV;
	cpu.Z  = cpu.a & temp1.si_;
}

void inline BMI()
{
	BRANCH(cpu.N & 0x80);
}

void inline BNE()
{
	BRANCH(cpu.Z);
}

void inline BPL()
{
	BRANCH(!(cpu.N & 0x80));
}

void inline BRK()
{
	ram[0x100|cpu.s]			= (cpu.pc+1)>>8;
	ram[0x100|(u8)(cpu.s-1)]	= (u8)(cpu.pc+1);
	ram[0x100|(u8)(cpu.s-2)]	= (cpu.N & cN) | (cpu.V?cV:0) | cR | cB | cpu.D | cpu.I | (cpu.Z ? 0: cZ) | cpu.C;
	cpu.s -=3;
	cpu.pc = pRead16(IRQVector);
	cpu.I  = cI;
}

void inline BVC()
{
	BRANCH(!cpu.V);
}

void inline BVS()
{
	BRANCH(cpu.V);
}

void inline CLC()
{
	cpu.C = 0;
}

void inline CLD()
{
	cpu.D = 0;
}

void inline CLI()
{
	cpu.I = 0;
}

void inline CLV()
{
	cpu.V = 0;
}

void inline CMP()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	temp1.si_	= cpu.a - Read6502();
	cpu.N		= cpu.Z = temp1.t16_.u.l;
	cpu.C		= temp1.u16_ < 256;
}

void inline CPX()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	temp1.si_	= cpu.x - Read6502();
	cpu.N		= cpu.Z = temp1.t16_.u.l;
	cpu.C		= temp1.u16_ < 256;
}

void inline CPY()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	temp1.si_	= cpu.y - Read6502();
	cpu.N		= cpu.Z = temp1.t16_.u.l;
	cpu.C		= temp1.u16_ < 256;
}

void inline DEC()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
		{
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
		}
	#endif
	temp1.t16_.u.l = Read6502();
	Write6502(temp1.t16_.u.l);
	cpu.N = cpu.Z = temp1.t16_.u.l -1;
	Write6502(temp1.t16_.u.l -1);
}

void inline DEX()
{
	cpu.N = cpu.Z = --cpu.x;
}

void inline DEY()
{
	cpu.N = cpu.Z = --cpu.y;
}

void inline EOR()
{
	cpu.N = cpu.Z = cpu.a ^= Read6502();
}

void inline INC()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
		{
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
		}
	#endif
	temp1.t16_.u.l = Read6502();
	Write6502(temp1.t16_.u.l);
	cpu.N = cpu.Z = (temp1.t16_.u.l+1);
	Write6502(temp1.t16_.u.l+1);
}

void inline INX()
{
	cpu.N = cpu.Z = ++cpu.x;
}

void inline INY()
{
	cpu.N = cpu.Z = ++cpu.y;
}

void inline JMP()
{
	#ifdef loggedCpu
		if (cpu.pc >= 0x8000)
		{
			loggedImage[ loggedOffset(cpu.pc)   ]		|= logged_opcode;
			loggedImage[ loggedOffset(cpu.pc+1) ]		|= logged_oprand;
			loggedImage[ loggedOffset(cpu.pc+2) ]		|= logged_oprand;
		}
	#endif
	cpu.pc = pRead16(cpu.pc+1);
	#ifdef loggedCpu
		if (cpu.pc >= 0x8000)
			loggedImage[ loggedOffset(cpu.pc) ]		|= logged_jumped;
	#endif
}

void inline JMPIND()
{
	tempVar temp1;
	#ifdef loggedCpu
		if (cpu.pc >= 0x8000)
		{
			loggedImage[ loggedOffset(cpu.pc)   ]		|= logged_opcode;
			loggedImage[ loggedOffset(cpu.pc+1) ]		|= logged_oprand;
			loggedImage[ loggedOffset(cpu.pc+2) ]		|= logged_oprand;
		}
	#endif
	temp1.si_	 = pRead16(cpu.pc+1);
	cpu.pc		 = pRead(temp1.si_);
	cpu.pc		|= (pbankRead[(temp1.si_)>>11][((temp1.si_) & 0x700) | (u8)(temp1.si_+1)])<<8;

	#ifdef loggedCpu
		if (cpu.pc >= 0x8000)
			loggedImage[ loggedOffset(cpu.pc) ]		|= logged_Indirectjump;
	#endif
}

void inline JSR()
{
	tempVar temp1;
	temp1.u16_ = pRead16(cpu.pc+1);
	ram[0x100|cpu.s] = (cpu.pc+2)>>8;
	ram[0x100|(u8)(cpu.s-1)] = (u8)(cpu.pc+2);
	cpu.s -= 2;
	cpu.pc = temp1.u16_;
}

void inline LDA()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	cpu.N = cpu.Z = cpu.a = Read6502();
}

void inline LDX()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	cpu.N = cpu.Z = cpu.x = Read6502();
}

void inline LDY()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	cpu.N = cpu.Z = cpu.y = Read6502();
}

void inline LSR()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	temp1.t16_.u.l = Read6502();
	Write6502(temp1.t16_.u.l);
	cpu.C		= temp1.t16_.u.l & 1;
	cpu.N = cpu.Z = (u8)(temp1.t16_.u.l>>1);
	Write6502(cpu.Z);
}

void inline LSRA()
{
	cpu.C = cpu.a & 1;
	cpu.N = cpu.Z = cpu.a >>= 1;
}

void inline NOP() {}

void inline ORA()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	cpu.N = cpu.Z = cpu.a |= Read6502();
}

void inline PHA()
{
	ram[0x100|cpu.s--] = cpu.a;
}

void inline PHP()
{
	ram[0x100|cpu.s--]  = (cpu.N & cN) | (cpu.V?cV:0) | cR | cB | cpu.D | cpu.I | (cpu.Z ? 0: cZ) | cpu.C;
}

void inline PLA()
{
	cpu.Z = cpu.N = cpu.a = ram[0x100|++cpu.s];
}

void inline PLP()
{
	cpu.s++;
	cpu.N = ram[0x100|cpu.s] & cN;
	cpu.V = ram[0x100|cpu.s] & cV;
	cpu.D = ram[0x100|cpu.s] & cD;
	cpu.I = ram[0x100|cpu.s] & cI;
	cpu.Z = !(ram[0x100|cpu.s] & cZ);
	cpu.C = ram[0x100|cpu.s] & cC;
}

void inline ROL()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
		{
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
		}
	#endif
	temp1.t16_.u.l= Read6502();
	Write6502(temp1.t16_.u.l);
	temp1.u16_	= temp1.t16_.u.l<<1 | cpu.C;
	cpu.Z		= cpu.N = temp1.t16_.u.l;
	cpu.C		= temp1.u16_ > 0xff;
	Write6502(temp1.t16_.u.l);
}

void inline ROLA()
{
	tempVar temp1;
	temp1.si_  = cpu.a<<1 | cpu.C;
	cpu.Z = cpu.N = cpu.a = temp1.t16_.u.l;
	cpu.C = temp1.si_ > 0xff;
}

void inline ROR()
{
	tempVar temp1;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
		{
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
		}
	#endif
	temp1.t16_.u.l = Read6502();
	Write6502(temp1.t16_.u.l);
	cpu.Z = cpu.N = (u8)(temp1.t16_.u.l >> 1 | (cpu.C ? 0x80: 0));
	cpu.C = temp1.t16_.u.l & 1;
	Write6502(cpu.Z);
}

void inline RORA()
{
	tempVar temp1;
	temp1.si_   = cpu.a>>1 | (cpu.C ? 0x80: 0);
	cpu.C  = cpu.a & 1;
	cpu.Z  = cpu.N = cpu.a = temp1.t16_.u.l;
}

void inline RTI()
{
	cpu.s;
	cpu.N = ram[0x100|(u8)(cpu.s+1)] & cN;
	cpu.V = ram[0x100|(u8)(cpu.s+1)] & cV;
	cpu.D = ram[0x100|(u8)(cpu.s+1)] & cD;
	cpu.I = ram[0x100|(u8)(cpu.s+1)] & cI;
	cpu.Z = !(ram[0x100|(u8)(cpu.s+1)] & cZ);
	cpu.C = ram[0x100|(u8)(cpu.s+1)] & cC;

	cpu.pc  = ram[0x100|(u8)(cpu.s+2)];
	cpu.pc |= ram[0x100|(u8)(cpu.s+3)]<<8;
	cpu.s +=3;
}


void inline RTS()
{
	cpu.pc  = ram[0x100|(u8)(cpu.s+1)];
	cpu.pc |= ram[0x100|(u8)(cpu.s+2)]<<8;
	cpu.pc++;
	cpu.s  += 2;
}

void inline SBC()
{
	tempVar temp1, temp2;
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 0);
	#endif
	temp1.si_	= Read6502();
	temp2.si_	= cpu.a - temp1.t16_.u.l - !cpu.C;
	cpu.V		= ( ((cpu.a ^ temp1.t16_.u.l) & 0x80) && ((cpu.a ^ temp2.t16_.u.l) & 0x80) );
	cpu.Z		= cpu.N = cpu.a = temp2.t16_.u.l;
	cpu.C		= temp2.u16_ < 256;
}

void inline SEC()
{
	cpu.C  = cC;
}

void inline SED()
{
	cpu.D  = cD;
}

void inline SEI()
{
	cpu.I  = cI;
}

void inline STA()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
	#endif
	Write6502(cpu.a);
}

void inline STX()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
	#endif
	Write6502(cpu.x);
}

void inline STY()
{
	#ifdef loggedCpu
		if ((memloc.si_ >= 0x8000) && (memloc.si_ <= 0xFFFF))
			loggedImage[ loggedOffset(memloc.si_) ]	|= getMode(rwMode, 1);
	#endif
	Write6502(cpu.y);
}

void inline TAX()
{
	cpu.N = cpu.Z = cpu.x = cpu.a;
}

void inline TAY()
{
	cpu.N = cpu.Z = cpu.y = cpu.a;
}

void inline TSX()
{
	cpu.N = cpu.Z = cpu.x = cpu.s;
}

void inline TXA()
{
	cpu.N = cpu.Z = cpu.a = cpu.x;
}

void inline TXS()
{
	cpu.s = cpu.x;
}

void inline TYA()
{
	cpu.N = cpu.Z = cpu.a = cpu.y;
}

void inline BRANCH(u8 b)
{
	tempVar temp1;
	if (b)
	{
		temp1.si_ = (cpu.pc+1)>>8;
		cpu.pc += (s8)(pRead(cpu.pc+1));
		cpu.pc += 2;
		if ((u8)(cpu.pc>>8) != (u8)temp1.si_)
			cpuCycles += 4;
		else
			cpuCycles += 3;
		
		#ifdef loggedCpu
			if (cpu.pc >= 0x8000)
				loggedImage[ loggedOffset(cpu.pc) ]		|= logged_branched;
		#endif
	}
	else
	{
		cpuCycles += 2;
		cpu.pc += 2;
	}
}

}
